(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Random = Package.random.Random;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var Accounts = Package['accounts-base'].Accounts;
var HTTP = Package.http.HTTP;
var HTTPInternals = Package.http.HTTPInternals;

/* Package-scope variables */
var SandstormDb, Packages, DevApps, UserActions, Grains, RoleAssignments, Contacts, Sessions, SignupKeys, ActivityStats, DeleteStats, FileTokens, ApiTokens, Notifications, StatsTokens, Misc, Settings, Migrations, StaticAssets, AssetUploadTokens, isDemoUser, isSignedUp, isSignedUpOrDemo, isUserOverQuota, isUserExcessivelyOverQuota, isAdmin, isAdminById, findAdminUserForToken, matchWildcardHost, makeWildcardHost, getWildcardOrigin, allowDevAccounts, roleAssignmentPattern, addStaticAsset, userPictureUrl, fetchPicture;

(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/db.js                                                                        //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                 // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                              // 2
// All rights reserved.                                                                               // 3
//                                                                                                    // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                    // 5
// you may not use this file except in compliance with the License.                                   // 6
// You may obtain a copy of the License at                                                            // 7
//                                                                                                    // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                       // 9
//                                                                                                    // 10
// Unless required by applicable law or agreed to in writing, software                                // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                  // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                           // 13
// See the License for the specific language governing permissions and                                // 14
// limitations under the License.                                                                     // 15
                                                                                                      // 16
// This file defines the database schema.                                                             // 17
                                                                                                      // 18
// Users = new Mongo.Collection("users");                                                             // 19
// The users collection is special and can be accessed through `Meteor.users`.                        // 20
// See https://docs.meteor.com/#/full/meteor_users. Entries in the users collection correspond to     // 21
// accounts. Each represents a capability store belonging to a human.                                 // 22
//                                                                                                    // 23
// Note that accounts are not the same thing as identities. An account may have multiple identities   // 24
// attached to it. Identities provide a way to authenticate the user and are also used to present     // 25
// globally unique and stable user indentifiers to grains via `Grain.UserInfo.userId`.                // 26
//                                                                                                    // 27
// Each entry in this collection contains:                                                            // 28
//   _id: Random.                                                                                     // 29
//   createdAt: Date when this entry was added to the collection.                                     // 30
//   lastActive: Date of the user's most recent interaction with this Sandstorm server.               // 31
//   profile: Object containing profile data editable by the user and visible to other users. May     // 32
//            include the following fields. Note that if any field is missing, the first fallback     // 33
//            is to check `services` for details provided by the identity provider (the details       // 34
//            of which differ per-provider). Only if that is also missing do we fall back to          // 35
//            defaults.                                                                               // 36
//       name: String containing the display name of the user. Default: first part of email.          // 37
//       handle: String containing the user's preferred handle. Default: first part of email.         // 38
//       picture: _id into the StaticAssets table for the user's picture. Default: identicon.         // 39
//       pronoun: One of "male", "female", "neutral", or "robot". Default: neutral.                   // 40
//   services: Object containing login and identity data used by Meteor authentication services.      // 41
//   isAdmin: Boolean indicating whether this user is allowed to access the Sandstorm admin panel.    // 42
//   signupKey: If this is an invited user, then this field contains their signup key.                // 43
//   signupNote: If the user was invited through a link, then this field contains the note that the   // 44
//               inviter admin attached to the key.                                                   // 45
//   signupEmail: If the user was invited by email, then this field contains the email address that   // 46
//                the invite was sent to.                                                             // 47
//   quota: Number of bytes this user is allowed to store.                                            // 48
//   storageUsage: Number of bytes this user is currently storing.                                    // 49
//   expires: Date when this user's account should be deleted. Only present for demo users.           // 50
//   isAppDemoUser: True if this is a demo user who arrived via an /appdemo/ link.                    // 51
                                                                                                      // 52
Packages = new Mongo.Collection("packages");                                                          // 53
// Packages which are installed or downloading.                                                       // 54
//                                                                                                    // 55
// Each contains:                                                                                     // 56
//   _id:  128-bit prefix of SHA-256 hash of spk file, hex-encoded.                                   // 57
//   status:  String.  One of "download", "verify", "unpack", "analyze", "ready", "failed", "delete"  // 58
//   progress:  Float.  -1 = N/A, 0-1 = fractional progress (e.g. download percentage),               // 59
//       >1 = download byte count.                                                                    // 60
//   error:  If status is "failed", error message string.                                             // 61
//   manifest:  If status is "ready", the package manifest.  See "Manifest" in grain.capnp.           // 62
//   appId:  If status is "ready", the application ID string.  Packages representing different        // 63
//       versions of the same app have the same appId.  The spk tool defines the app ID format        // 64
//       and can cryptographically verify that a package belongs to a particular app ID.              // 65
                                                                                                      // 66
DevApps = new Mongo.Collection("devapps");                                                            // 67
// List of applications currently made available via the dev tools running on the local machine.      // 68
// This is normally empty; the only time it is non-empty is when a developer is using the spk tool    // 69
// on the local machine to publish an under-development app to this server. That should only ever     // 70
// happen on developers' desktop machines.                                                            // 71
//                                                                                                    // 72
// While a dev app is published, it automatically appears as installed by every user of the server,   // 73
// and it overrides all packages with the same application ID. If any instances of those packages     // 74
// are currently open, they are killed and reset on publish.                                          // 75
//                                                                                                    // 76
// When the dev tool disconnects, the app is automatically unpublished, and any open instances        // 77
// are again killed and refreshed.                                                                    // 78
//                                                                                                    // 79
// Each contains:                                                                                     // 80
//   _id:  The application ID string (as with Packages.appId).                                        // 81
//   packageId:  The directory name where the dev package is mounted.                                 // 82
//   timestamp:  Time when the package was last updated. If this changes while the package is         // 83
//     published, all running instances are reset. This is used e.g. to reset the app each time       // 84
//     changes are made to the source code.                                                           // 85
//   manifest:  The app's manifest, as with Packages.manifest.                                        // 86
                                                                                                      // 87
UserActions = new Mongo.Collection("userActions");                                                    // 88
// List of actions that each user has installed which create new grains.  Each app may install        // 89
// some number of actions (usually, one).                                                             // 90
//                                                                                                    // 91
// Each contains:                                                                                     // 92
//   _id:  random                                                                                     // 93
//   userId:  User who has installed this action.                                                     // 94
//   packageId:  Package used to run this action.                                                     // 95
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.               // 96
//   appTitle:  Same as Packages.findOne(packageId).manifest.appTitle.defaultText; denormalized so    // 97
//       that clients can access it without subscribing to the Packages collection.                   // 98
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for           // 99
//       searchability.                                                                               // 100
//   appMarketingVersion:  Human-readable presentation of the app version, e.g. "2.9.17"              // 101
//   title:  Human-readable title for this action, e.g. "New Spreadsheet".                            // 102
//   command:  Manifest.Command to run this action (see package.capnp).                               // 103
                                                                                                      // 104
Grains = new Mongo.Collection("grains");                                                              // 105
// Grains belonging to users.                                                                         // 106
//                                                                                                    // 107
// Each contains:                                                                                     // 108
//   _id:  random                                                                                     // 109
//   packageId:  _id of the package of which this grain is an instance.                               // 110
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.               // 111
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for           // 112
//       searchability.                                                                               // 113
//   userId:  User who owns this grain.                                                               // 114
//   title:  Human-readable string title, as chosen by the user.                                      // 115
//   lastUsed:  Date when the grain was last used by a user.                                          // 116
//   private: If true, then knowledge of `_id` does not suffice to open this grain.                   // 117
//   cachedViewInfo: The JSON-encoded result of `UiView.getViewInfo()`, cached from the most recent   // 118
//                   time a session to this grain was opened.                                         // 119
//                                                                                                    // 120
// The following fields *might* also exist. These are temporary hacks used to implement e-mail and    // 121
// web publishing functionality without powerbox support; they will be replaced once the powerbox     // 122
// is implemented.                                                                                    // 123
//   publicId:  An id used to publicly identify this grain. Used e.g. to route incoming e-mail and    // 124
//       web publishing. This field is initialized when first requested by the app.                   // 125
                                                                                                      // 126
RoleAssignments = new Mongo.Collection("roleAssignments");                                            // 127
// *OBSOLETE* Before `user` was a variant of ApiTokenOwner, this collection was used to store edges   // 128
// in the permissions sharing graph. This functionality has been subsumed by the ApiTokens            // 129
// collection.                                                                                        // 130
                                                                                                      // 131
Contacts = new Mongo.Collection("contacts");                                                          // 132
// Edges in the social graph.                                                                         // 133
//                                                                                                    // 134
// If Alice has Bob as a contact, then she is allowed to see Bob's profile information and Bob        // 135
// will show up in her user-picker UI for actions like share-by-identity.                             // 136
//                                                                                                    // 137
// Contacts are not symmetric. Bob might be one of Alice's contacts even if Alice is not one of       // 138
// Bob's.                                                                                             // 139
//                                                                                                    // 140
// Each contains:                                                                                     // 141
//   _id: random                                                                                      // 142
//   ownerId: The `_id` of the user who owns this contact.                                            // 143
//   userId:  The `_id` of the contacted user.                                                        // 144
//   petname: Human-readable label chosen by and only visible to the owner. Uniquely identifies       // 145
//            the contact to the owner.                                                               // 146
//   created: Date when this contact was created.                                                     // 147
                                                                                                      // 148
Sessions = new Mongo.Collection("sessions");                                                          // 149
// UI sessions open to particular grains.  A new session is created each time a user opens a grain.   // 150
//                                                                                                    // 151
// Each contains:                                                                                     // 152
//   _id:  random                                                                                     // 153
//   grainId:  _id of the grain to which this session is connected.                                   // 154
//   hostId: ID part of the hostname from which this grain is being served. I.e. this replaces the    // 155
//       '*' in WILDCARD_HOST.                                                                        // 156
//   timestamp:  Time of last keep-alive message to this session.  Sessions time out after some       // 157
//       period.                                                                                      // 158
//   userId:  User who owns this session.                                                             // 159
//   hashedToken: If the session is owned by an anonymous user, the _id of the entry in ApiTokens     // 160
//       that was used to open it. Note that for old-style sharing (i.e. when !grain.private),        // 161
//       anonymous users can get access without an API token and so neither userId nor hashedToken    // 162
//       are present.                                                                                 // 163
//   powerboxView: If present, this is a view that should be presented as part of a powerbox          // 164
//       interaction.                                                                                 // 165
//     offer: The webkey that corresponds to cap that was passed to the `offer` RPC.                  // 166
                                                                                                      // 167
SignupKeys = new Mongo.Collection("signupKeys");                                                      // 168
// Invite keys which may be used by users to get access to Sandstorm.                                 // 169
//                                                                                                    // 170
// Each contains:                                                                                     // 171
//   _id:  random                                                                                     // 172
//   used:  Boolean indicating whether this key has already been consumed.                            // 173
//   note:  Text note assigned when creating key, to keep track of e.g. whom the key was for.         // 174
//   email: If this key was sent as an email invite, the email address to which it was sent.          // 175
//   quota: If present, the storage quota to assign to the user claiming this invite.                 // 176
                                                                                                      // 177
ActivityStats = new Mongo.Collection("activityStats");                                                // 178
// Contains usage statistics taken on a regular interval. Each entry is a data point.                 // 179
//                                                                                                    // 180
// Each contains:                                                                                     // 181
//   timestamp: Date when measurements were taken.                                                    // 182
//   daily: Contains stats counts pertaining to the last day before the sample time.                  // 183
//   weekly: Contains stats counts pertaining to the last seven days before the sample time.          // 184
//   monthly: Contains stats counts pertaining to the last thirty days before the timestamp.          // 185
//                                                                                                    // 186
// Each of daily, weekly, and monthly contains:                                                       // 187
//   activeUsers: The number of unique users who have used a grain on the server in the time          // 188
//       interval. Only counts logged-in users.                                                       // 189
//   activeGrains: The number of unique grains that have been used in the time interval.              // 190
                                                                                                      // 191
DeleteStats = new Mongo.Collection("deleteStats");                                                    // 192
// Contains records of objects that were deleted, for stat-keeping purposes.                          // 193
//                                                                                                    // 194
// Each contains:                                                                                     // 195
//   type: "grain" or "user" or "appDemoUser"                                                         // 196
//   lastActive: Date of the user's or grain's last activity.                                         // 197
                                                                                                      // 198
FileTokens = new Mongo.Collection("fileTokens");                                                      // 199
// Tokens corresponding to backup files that are currently stored on the server. A user receives      // 200
// a token when they create a backup file (either by uploading it, or by backing up one of their      // 201
// grains) and may use the token to read the file (either to download it, or to restore a new         // 202
// grain from it).                                                                                    // 203
//                                                                                                    // 204
// Each contains:                                                                                     // 205
//   _id:       The unguessable token string.                                                         // 206
//   name:      Suggested filename.                                                                   // 207
//   timestamp: File creation time. Used to figure out when the token and file should be wiped.       // 208
                                                                                                      // 209
ApiTokens = new Mongo.Collection("apiTokens");                                                        // 210
// Access tokens for APIs exported by apps.                                                           // 211
//                                                                                                    // 212
// Originally API tokens were only used by external users through the HTTP API endpoint. However,     // 213
// now they are also used to implement SturdyRefs, not just held by external users, but also when     // 214
// an app holds a SturdyRef to another app within the same server. See the various `save()`,          // 215
// `restore()`, and `drop()` methods in `grain.capnp` (on `SansdtormApi`, `AppPersistent`, and        // 216
// `MainView`) -- the fields of type `Data` are API tokens.                                           // 217
//                                                                                                    // 218
// Each contains:                                                                                     // 219
//   _id:       A SHA-256 hash of the token.                                                          // 220
//   grainId:   The grain servicing this API. (Not present if the API isn't serviced by a grain.)     // 221
//   userId:    For UiView capabilities, this is the user for whom the view is attenuated. That       // 222
//              is, the UiView's newSession() method will intersect the requested permissions with    // 223
//              this user's permissions before forwarding on to the underlying app. If `userId` is    // 224
//              not present, then no user attenuation is applied, i.e. this is a raw UiView as        // 225
//              implemented by the app. (The `roleAssignment` field, below, may still apply.)         // 226
//              For non-UiView capabilities, `userId` is never present.                               // 227
//              Note that this is NOT the user against whom the `requiredPermissions` parameter of    // 228
//              `SandstormApi.restore()` is checked; that would be `owner.grain.introducerUser`.      // 229
//   userInfo:  *DEPRECATED* For API tokens created by the app through HackSessionContext, the        // 230
//              UserInfo struct that should be passed to `newSession()` when exercising this token,   // 231
//              in decoded (JS object) format. This is a temporary hack. `userId` is never present    // 232
//              when `userInfo` is present.                                                           // 233
//   roleAssignment: If this API token represents a UiView, this field contains a JSON-encoded        // 234
//              Grain.ViewSharingLink.RoleAssignment representing the permissions it carries. These   // 235
//              permissions will be intersected with those held by `userId` when the view is opened.  // 236
//   forSharing: If true, requests sent to the HTTP API endpoint with this token will be treated as   // 237
//              anonymous rather than as directly associated with `userId`. This has no effect on     // 238
//              the permissions granted.                                                              // 239
//   objectId:  If present, this token represents an arbitrary Cap'n Proto capability exported by     // 240
//              the app or its supervisor (whereas without this it strictly represents UiView).       // 241
//              sturdyRef is the JSON-encoded SupervisorObjectId (defined in `supervisor.capnp`).     // 242
//              Note that if the SupervisorObjectId contains an AppObjectId, that field is            // 243
//              treated as type AnyPointer, and so encoded as a raw Cap'n Proto message.              // 244
//   frontendRef: If present, this token actually refers to an object implemented by the front-end,   // 245
//              not a particular grain. (`grainId` and `userId` are not set.) This is an object       // 246
//              containing exactly one of the following fields:                                       // 247
//       notificationHandle: A `Handle` for an ongoing notification, as returned by                   // 248
//                           `NotificationTarget.addOngoing`. The value is an `_id` from the          // 249
//                           `Notifications` collection.                                              // 250
//       ipNetwork: An IpNetwork capability that is implemented by the frontend. Eventually, this     // 251
//                  will be moved out of the frontend and into the backend, but we'll migrate the     // 252
//                  database when that happens. This field contains the boolean true to signify that  // 253
//                  it has been set.                                                                  // 254
//       ipInterface: Ditto IpNetwork, except it's an IpInterface.                                    // 255
//   parentToken: If present, then this token represents exactly the capability represented by        // 256
//              the ApiToken with _id = parentToken, except possibly (if it is a UiView) attenuated   // 257
//              by `roleAssignment` (if present). To facilitate permissions computations, if the      // 258
//              capability is a UiView, then `grainId` is set to the backing grain and `userId` is    // 259
//              set to the user who shared the view. None of `userInfo`, `objectId`, or               // 260
//              `frontendRef` are present when `parentToken` is present.                              // 261
//   petname:   Human-readable label for this access token, useful for identifying tokens for         // 262
//              revocation. This should be displayed when visualizing incoming capabilities to        // 263
//              the grain identified by `grainId`.                                                    // 264
//   created:   Date when this token was created.                                                     // 265
//   revoked:   If true, then this sturdyref has been revoked and can no longer be restored. It may   // 266
//              become un-revoked in the future.                                                      // 267
//   expires:   Optional expiration Date. If undefined, the token does not expire.                    // 268
//   owner:     A `ApiTokenRefOwner` (defined in `supervisor.capnp`, stored as a JSON object)         // 269
//              as passed to the `save()` call that created this token. If not present, treat         // 270
//              as `webkey` (the default for `ApiTokenOwner`).                                        // 271
//   expiresIfUnused:                                                                                 // 272
//              Optional Date after which the token, if it has not been used yet, expires.            // 273
//              This field should be cleared on a token's first use.                                  // 274
//   requirements: List of conditions which must hold for this token to be considered valid.          // 275
//              Semantically, this list specifies the powers which were *used* to originally          // 276
//              create the token. If any condition in the list becomes untrue, then the token must    // 277
//              be considered revoked, and all live refs and sturdy refs obtained transitively        // 278
//              through it must also become revoked. Each item is the JSON serialization of the       // 279
//              `MembraneRequirement` structure defined in `supervisor.capnp`.                        // 280
//                                                                                                    // 281
// It is important to note that a token's owner and provider are independent from each other. To      // 282
// illustrate, here is an approximate definition of ApiToken in pseudo Cap'n Proto schema language:   // 283
//                                                                                                    // 284
// struct ApiToken {                                                                                  // 285
//   owner :ApiTokenOwner;                                                                            // 286
//   provider :union {                                                                                // 287
//     grain :group {                                                                                 // 288
//       grainId :Text;                                                                               // 289
//       union {                                                                                      // 290
//         uiView :group {                                                                            // 291
//           userId :Text;                                                                            // 292
//           roleAssignment :RoleAssignment;                                                          // 293
//           forSharing :Bool;                                                                        // 294
//         }                                                                                          // 295
//         objectId :SupervisorObjectId;                                                              // 296
//       }                                                                                            // 297
//     }                                                                                              // 298
//     frontendRef :union {                                                                           // 299
//        notificationHandle :Text;                                                                   // 300
//        ipNetwork :Bool;                                                                            // 301
//        ipInterface :Bool;                                                                          // 302
//     }                                                                                              // 303
//     child :group {                                                                                 // 304
//       parentToken :Text;                                                                           // 305
//       union {                                                                                      // 306
//         uiView :group {                                                                            // 307
//           grainId :Text;                                                                           // 308
//           userId :Text;                                                                            // 309
//           roleAssignment :RoleAssignment = (allAccess = ());                                       // 310
//         }                                                                                          // 311
//         other :Void;                                                                               // 312
//       }                                                                                            // 313
//     }                                                                                              // 314
//   }                                                                                                // 315
//   requirements: List(Supervisor.MembraneRequirement);                                              // 316
//   ...                                                                                              // 317
// }                                                                                                  // 318
                                                                                                      // 319
Notifications = new Mongo.Collection("notifications");                                                // 320
// Notifications for a user.                                                                          // 321
//                                                                                                    // 322
// Each contains:                                                                                     // 323
//   _id:          random                                                                             // 324
//   ongoing:      If present, this is an ongoing notification, and this field contains an            // 325
//                 ApiToken referencing the `OngoingNotification` capability.                         // 326
//   grainId:      The grain originating this notification, if any.                                   // 327
//   userId:       The user receiving the notification.                                               // 328
//   text:         The JSON-ified LocalizedText to display in the notification.                       // 329
//   isUnread:     Boolean indicating if this notification is unread.                                 // 330
//   timestamp:    Date when this notification was last updated                                       // 331
                                                                                                      // 332
StatsTokens = new Mongo.Collection("statsTokens");                                                    // 333
// Access tokens for the Stats collection                                                             // 334
//                                                                                                    // 335
// These tokens are used for accessing the ActivityStats collection remotely                          // 336
// (ie. from a dashboard webapp)                                                                      // 337
//                                                                                                    // 338
// Each contains:                                                                                     // 339
//   _id:       The token. At least 128 bits entropy (Random.id(22)).                                 // 340
                                                                                                      // 341
Misc = new Mongo.Collection("misc");                                                                  // 342
// Miscellaneous configuration and other settings                                                     // 343
//                                                                                                    // 344
// This table is currently only used for persisting BASE_URL from one session to the next,            // 345
// but in general any miscellaneous settings should go in here                                        // 346
//                                                                                                    // 347
// Each contains:                                                                                     // 348
//   _id:       The name of the setting. eg. "BASE_URL"                                               // 349
//   value:     The value of the setting.                                                             // 350
                                                                                                      // 351
Settings = new Mongo.Collection("settings");                                                          // 352
// Settings for this Sandstorm instance go here. They are configured through the adminSettings        // 353
// route. This collection differs from misc in that this collection is completely user controlled.    // 354
//                                                                                                    // 355
// Each contains:                                                                                     // 356
//   _id:       The name of the setting. eg. "MAIL_URL"                                               // 357
//   value:     The value of the setting.                                                             // 358
//   potentially other fields that are unique to the setting                                          // 359
                                                                                                      // 360
Migrations = new Mongo.Collection("migrations");                                                      // 361
// This table tracks which migrations we have applied to this instance.                               // 362
// It contains a single entry:                                                                        // 363
//   _id:       "migrations_applied"                                                                  // 364
//   value:     The number of migrations this instance has successfully completed.                    // 365
                                                                                                      // 366
StaticAssets = new Mongo.Collection("staticAssets");                                                  // 367
// Collection of static assets served up from the Sandstorm server's "static" host. We only           // 368
// support relatively small assets: under 1MB each.                                                   // 369
//                                                                                                    // 370
// Each contains:                                                                                     // 371
//   _id:       Random ID; will be used in the URL.                                                   // 372
//   hash:      A SHA-256 hash of the data, used to de-dupe.                                          // 373
//   mimeType:  MIME type of the asset, suitable for Content-Type header.                             // 374
//   encoding:  Either "gzip" or not present, suitable for Content-Encoding header.                   // 375
//   content:   The asset content (byte buffer).                                                      // 376
//   refcount:  Number of places where this asset's ID appears in the database. Since Mongo doesn't   // 377
//       have transactions, this needs to bias towards over-counting; a backup GC could be used       // 378
//       to catch leaked assets, although it's probably not a big deal in practice.                   // 379
                                                                                                      // 380
AssetUploadTokens = new Mongo.Collection("assetUploadTokens");                                        // 381
// Collection of tokens representing a single-use permission to upload an asset, such as a new        // 382
// profile picture.                                                                                   // 383
//                                                                                                    // 384
// Each contains:                                                                                     // 385
//   _id:       Random ID.                                                                            // 386
//   purpose:   Contains one of the following, indicating how the asset is to be used:                // 387
//       profilePicture: Indicates that the upload is a new profile picture. Contains fields:         // 388
//           userId: User whose picture shall be replaced.                                            // 389
//           identityId: Which of the user's identities shall be updated.                             // 390
//   expires:   Time when this token will go away if unused.                                          // 391
                                                                                                      // 392
if (Meteor.isServer) {                                                                                // 393
  Meteor.publish("credentials", function () {                                                         // 394
    // Data needed for isSignedUp() and isAdmin() to work.                                            // 395
                                                                                                      // 396
    if (this.userId) {                                                                                // 397
      return Meteor.users.find({_id: this.userId},                                                    // 398
          {fields: {signupKey: 1, isAdmin: 1, expires: 1, quota: 1, storageUsage: 1}});               // 399
    } else {                                                                                          // 400
      return [];                                                                                      // 401
    }                                                                                                 // 402
  });                                                                                                 // 403
}                                                                                                     // 404
                                                                                                      // 405
isDemoUser = function() {                                                                             // 406
  // Returns true if this is a demo user.                                                             // 407
                                                                                                      // 408
  var user = Meteor.user();                                                                           // 409
  if (user && user.expires) {                                                                         // 410
    return true;                                                                                      // 411
  } else {                                                                                            // 412
    return false;                                                                                     // 413
  }                                                                                                   // 414
}                                                                                                     // 415
                                                                                                      // 416
isSignedUp = function() {                                                                             // 417
  // Returns true if the user has presented an invite key.                                            // 418
                                                                                                      // 419
  var user = Meteor.user();                                                                           // 420
  if (user && user.signupKey) {                                                                       // 421
    return true;                                                                                      // 422
  } else {                                                                                            // 423
    return false;                                                                                     // 424
  }                                                                                                   // 425
}                                                                                                     // 426
                                                                                                      // 427
isSignedUpOrDemo = function () {                                                                      // 428
  var user = Meteor.user();                                                                           // 429
  if (user && (user.signupKey || user.expires)) {                                                     // 430
    return true;                                                                                      // 431
  } else {                                                                                            // 432
    return false;                                                                                     // 433
  }                                                                                                   // 434
}                                                                                                     // 435
                                                                                                      // 436
isUserOverQuota = function (user) {                                                                   // 437
  return Meteor.settings.public.quotaEnabled &&                                                       // 438
         typeof user.quota === "number" &&                                                            // 439
         user.storageUsage && user.storageUsage >= user.quota;                                        // 440
}                                                                                                     // 441
                                                                                                      // 442
isUserExcessivelyOverQuota = function (user) {                                                        // 443
  return Meteor.settings.public.quotaEnabled &&                                                       // 444
         typeof user.quota === "number" &&                                                            // 445
         user.storageUsage && user.storageUsage >= user.quota * 1.2;                                  // 446
}                                                                                                     // 447
                                                                                                      // 448
isAdmin = function() {                                                                                // 449
  // Returns true if the user is the administrator.                                                   // 450
                                                                                                      // 451
  var user = Meteor.user();                                                                           // 452
  if (user && user.isAdmin) {                                                                         // 453
    return true;                                                                                      // 454
  } else {                                                                                            // 455
    return false;                                                                                     // 456
  }                                                                                                   // 457
}                                                                                                     // 458
                                                                                                      // 459
isAdminById = function(id) {                                                                          // 460
  // Returns true if the user's id is the administrator.                                              // 461
                                                                                                      // 462
  var user = Meteor.users.findOne({_id: id}, {fields: {isAdmin: 1}})                                  // 463
  if (user && user.isAdmin) {                                                                         // 464
    return true;                                                                                      // 465
  } else {                                                                                            // 466
    return false;                                                                                     // 467
  }                                                                                                   // 468
}                                                                                                     // 469
                                                                                                      // 470
findAdminUserForToken = function (token) {                                                            // 471
  if (!token.requirements) {                                                                          // 472
    return;                                                                                           // 473
  }                                                                                                   // 474
  var requirements = token.requirements.filter(function (requirement) {                               // 475
    return "userIsAdmin" in requirement;                                                              // 476
  });                                                                                                 // 477
                                                                                                      // 478
  if (requirements.length > 1) {                                                                      // 479
    return;                                                                                           // 480
  }                                                                                                   // 481
  if (requirements.length === 0) {                                                                    // 482
    return;                                                                                           // 483
  }                                                                                                   // 484
  return requirements[0].userIsAdmin;                                                                 // 485
};                                                                                                    // 486
                                                                                                      // 487
var wildcardHost = Meteor.settings.public.wildcardHost.toLowerCase().split("*");                      // 488
                                                                                                      // 489
if (wildcardHost.length != 2) {                                                                       // 490
  throw new Error("Wildcard host must contain exactly one asterisk.");                                // 491
}                                                                                                     // 492
                                                                                                      // 493
matchWildcardHost = function(host) {                                                                  // 494
  // See if the hostname is a member of our wildcard. If so, extract the ID.                          // 495
                                                                                                      // 496
  var prefix = wildcardHost[0];                                                                       // 497
  var suffix = wildcardHost[1];                                                                       // 498
  if (host.lastIndexOf(prefix, 0) >= 0 &&                                                             // 499
      host.indexOf(suffix, -suffix.length) >= 0 &&                                                    // 500
      host.length >= prefix.length + suffix.length) {                                                 // 501
    var id = host.slice(prefix.length, -suffix.length);                                               // 502
    if (id.match(/^[a-z0-9]*$/)) {                                                                    // 503
      return id;                                                                                      // 504
    }                                                                                                 // 505
  }                                                                                                   // 506
                                                                                                      // 507
  return null;                                                                                        // 508
}                                                                                                     // 509
                                                                                                      // 510
makeWildcardHost = function (id) {                                                                    // 511
  return wildcardHost[0] + id + wildcardHost[1];                                                      // 512
}                                                                                                     // 513
                                                                                                      // 514
if (Meteor.isServer) {                                                                                // 515
  var Url = Npm.require("url");                                                                       // 516
  getWildcardOrigin = function () {                                                                   // 517
    // The wildcard URL can be something like "foo-*-bar.example.com", but sometimes when we're       // 518
    // trying to specify a pattern matching hostnames (say, a Content-Security-Policy directive),     // 519
    // an astrisk is only allowed as the first character and must be followed by a period. So we need // 520
    // "*.example.com" instead -- which matches more than we actually want, but is the best we can    // 521
    // really do. We also add the protocol to the front (again, that's what CSP wants).               // 522
                                                                                                      // 523
    // TODO(cleanup): `protocol` is computed in other files, like proxy.js. Put it somewhere common.  // 524
    var protocol = Url.parse(process.env.ROOT_URL).protocol;                                          // 525
                                                                                                      // 526
    var dotPos = wildcardHost[1].indexOf(".");                                                        // 527
    if (dotPos < 0) {                                                                                 // 528
      return protocol + "//*";                                                                        // 529
    } else {                                                                                          // 530
      return protocol + "//*" + wildcardHost[1].slice(dotPos);                                        // 531
    }                                                                                                 // 532
  }                                                                                                   // 533
}                                                                                                     // 534
                                                                                                      // 535
allowDevAccounts = function () {                                                                      // 536
  var setting = Settings.findOne({_id: "devAccounts"});                                               // 537
  if (setting) {                                                                                      // 538
    return setting.value;                                                                             // 539
  } else {                                                                                            // 540
    return Meteor.settings && Meteor.settings.public &&                                               // 541
           Meteor.settings.public.allowDevAccounts;                                                   // 542
  }                                                                                                   // 543
};                                                                                                    // 544
                                                                                                      // 545
roleAssignmentPattern = {                                                                             // 546
  none : Match.Optional(null),                                                                        // 547
  allAccess: Match.Optional(null),                                                                    // 548
  roleId: Match.Optional(Match.Integer),                                                              // 549
  addPermissions: Match.Optional([Boolean]),                                                          // 550
  removePermissions: Match.Optional([Boolean]),                                                       // 551
};                                                                                                    // 552
                                                                                                      // 553
SandstormDb = function () {                                                                           // 554
  this.collections = {                                                                                // 555
    // Direct access to underlying collections. DEPRECATED.                                           // 556
    //                                                                                                // 557
    // TODO(cleanup): Over time, we will provide methods covering each supported query and remove     // 558
    //   direct access to the collections.                                                            // 559
                                                                                                      // 560
    packages: Packages,                                                                               // 561
    devApps: DevApps,                                                                                 // 562
    userActions: UserActions,                                                                         // 563
    grains: Grains,                                                                                   // 564
    contacts: Contacts,                                                                               // 565
    sessions: Sessions,                                                                               // 566
    signupKeys: SignupKeys,                                                                           // 567
    activityStats: ActivityStats,                                                                     // 568
    deleteStats: DeleteStats,                                                                         // 569
    fileTokens: FileTokens,                                                                           // 570
    apiTokens: ApiTokens,                                                                             // 571
    notifications: Notifications,                                                                     // 572
    statsTokens: StatsTokens,                                                                         // 573
    misc: Misc,                                                                                       // 574
    settings: Settings,                                                                               // 575
                                                                                                      // 576
    // Intentionally omitted:                                                                         // 577
    // - Migrations, since it's used only within this package.                                        // 578
    // - RoleAssignments, since it is deprecated and only used by the migration that eliminated it.   // 579
  };                                                                                                  // 580
};                                                                                                    // 581
                                                                                                      // 582
// TODO(cleanup): These methods should not be defined freestanding and should use collection          // 583
//   objects created in SandstormDb's constructor rather than globals.                                // 584
                                                                                                      // 585
_.extend(SandstormDb.prototype, {                                                                     // 586
  isDemoUser: isDemoUser,                                                                             // 587
  isSignedUp: isSignedUp,                                                                             // 588
  isSignedUpOrDemo: isSignedUpOrDemo,                                                                 // 589
  isUserOverQuota: isUserOverQuota,                                                                   // 590
  isUserExcessivelyOverQuota: isUserExcessivelyOverQuota,                                             // 591
  isAdmin: isAdmin,                                                                                   // 592
  isAdminById: isAdminById,                                                                           // 593
  findAdminUserForToken: findAdminUserForToken,                                                       // 594
  matchWildcardHost: matchWildcardHost,                                                               // 595
  makeWildcardHost: makeWildcardHost,                                                                 // 596
  allowDevAccounts: allowDevAccounts,                                                                 // 597
  roleAssignmentPattern: roleAssignmentPattern,                                                       // 598
});                                                                                                   // 599
                                                                                                      // 600
if (Meteor.isServer) {                                                                                // 601
  SandstormDb.prototype.getWildcardOrigin = getWildcardOrigin;                                        // 602
}                                                                                                     // 603
                                                                                                      // 604
// =======================================================================================            // 605
// Below this point are newly-written or refactored functions.                                        // 606
                                                                                                      // 607
if (Meteor.isServer) {                                                                                // 608
  var Crypto = Npm.require("crypto");                                                                 // 609
  var ContentType = Npm.require("content-type");                                                      // 610
  var Zlib = Npm.require("zlib");                                                                     // 611
                                                                                                      // 612
  // TODO(cleanup): Node 0.12 has a `gzipSync` but 0.10 (which Meteor still uses) does not.           // 613
  var gzipSync = Meteor.wrapAsync(Zlib.gzip, Zlib);                                                   // 614
                                                                                                      // 615
  var BufferSmallerThan = function (limit) {                                                          // 616
    return Match.Where(function (buf) {                                                               // 617
      check(buf, Buffer);                                                                             // 618
      return buf.length < limit;                                                                      // 619
    });                                                                                               // 620
  }                                                                                                   // 621
                                                                                                      // 622
  var DatabaseId = Match.Where(function (s) {                                                         // 623
    check(s, String);                                                                                 // 624
    return !!s.match(/^[a-zA-Z0-9_]+$/);                                                              // 625
  });                                                                                                 // 626
                                                                                                      // 627
  addStaticAsset = function (metadata, content) {                                                     // 628
    // Add a new static asset to the database. If `content` is a string rather than a buffer, it      // 629
    // will be automatically gzipped before storage; do not specify metadata.encoding in this case.   // 630
                                                                                                      // 631
    if (typeof content === "string" && !metadata.encoding) {                                          // 632
      content = gzipSync(new Buffer(content, "utf8"));                                                // 633
      metadata.encoding = "gzip";                                                                     // 634
    }                                                                                                 // 635
                                                                                                      // 636
    check(metadata, {                                                                                 // 637
      mimeType: String,                                                                               // 638
      encoding: Match.Optional("gzip")                                                                // 639
    });                                                                                               // 640
    check(content, BufferSmallerThan(1 << 20));                                                       // 641
                                                                                                      // 642
    // Validate content type.                                                                         // 643
    metadata.mimeType = ContentType.format(ContentType.parse(metadata.mimeType));                     // 644
                                                                                                      // 645
    var hasher = Crypto.createHash("sha256");                                                         // 646
    hasher.update(metadata.mimeType + "\n" + metadata.encoding + "\n", "utf8");                       // 647
    hasher.update(content);                                                                           // 648
    var hash = hasher.digest("base64");                                                               // 649
                                                                                                      // 650
    var existing = StaticAssets.findAndModify({                                                       // 651
      query: {hash: hash, refcount: {$gte: 1}},                                                       // 652
      update: {$inc: {refcount: 1}},                                                                  // 653
      fields: {_id: 1, refcount: 1},                                                                  // 654
    });                                                                                               // 655
    if (existing) {                                                                                   // 656
      return existing._id;                                                                            // 657
    }                                                                                                 // 658
                                                                                                      // 659
    return StaticAssets.insert(_.extend({                                                             // 660
      hash: hash,                                                                                     // 661
      content: content,                                                                               // 662
      refcount: 1                                                                                     // 663
    }, metadata));                                                                                    // 664
  }                                                                                                   // 665
                                                                                                      // 666
  SandstormDb.prototype.addStaticAsset = addStaticAsset;                                              // 667
                                                                                                      // 668
  SandstormDb.prototype.refStaticAsset = function (id) {                                              // 669
    // Increment the refcount on an existing static asset.                                            // 670
    //                                                                                                // 671
    // You must call this BEFORE adding the new reference to the DB, in case of failure between       // 672
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big      // 673
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                        // 674
                                                                                                      // 675
    check(id, String);                                                                                // 676
                                                                                                      // 677
    var existing = StaticAssets.findAndModify({                                                       // 678
      query: {hash: hash},                                                                            // 679
      update: {$inc: {refcount: 1}},                                                                  // 680
      fields: {_id: 1, refcount: 1},                                                                  // 681
    });                                                                                               // 682
    if (!existing) {                                                                                  // 683
      throw new Error("refStaticAsset() called on asset that doesn't exist");                         // 684
    }                                                                                                 // 685
  }                                                                                                   // 686
                                                                                                      // 687
  SandstormDb.prototype.unrefStaticAsset = function (id) {                                            // 688
    // Decrement refcount on a static asset and delete if it has reached zero.                        // 689
    //                                                                                                // 690
    // You must call this AFTER removing the reference from the DB, in case of failure between        // 691
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big      // 692
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                        // 693
                                                                                                      // 694
    check(id, String);                                                                                // 695
                                                                                                      // 696
    var existing = StaticAssets.findAndModify({                                                       // 697
      query: {_id: id},                                                                               // 698
      update: {$inc: {refcount: -1}},                                                                 // 699
      fields: {_id: 1, refcount: 1},                                                                  // 700
      new: true,                                                                                      // 701
    });                                                                                               // 702
    if (!existing) {                                                                                  // 703
      console.error(new Error("unrefStaticAsset() called on asset that doesn't exist").stack);        // 704
    } else if (existing.refcount <= 0) {                                                              // 705
      StaticAssets.remove({_id: existing._id});                                                       // 706
    }                                                                                                 // 707
  }                                                                                                   // 708
                                                                                                      // 709
  SandstormDb.prototype.getStaticAsset = function (id) {                                              // 710
    // Get a static asset's mimeType, encoding, and raw content.                                      // 711
                                                                                                      // 712
    check(id, String);                                                                                // 713
                                                                                                      // 714
    var asset = StaticAssets.findOne(id, {fields: {_id: 0, mimeType: 1, encoding: 1, content: 1}});   // 715
    if (asset) {                                                                                      // 716
      // TODO(perf): Mongo converts buffers to something else. Figure out a way to avoid a copy       // 717
      //   here.                                                                                      // 718
      asset.content = new Buffer(asset.content);                                                      // 719
    }                                                                                                 // 720
    return asset;                                                                                     // 721
  }                                                                                                   // 722
                                                                                                      // 723
  SandstormDb.prototype.newAssetUpload = function (purpose) {                                         // 724
    check(purpose, {profilePicture: {userId: DatabaseId, identityId: Match.Optional(DatabaseId)}});   // 725
                                                                                                      // 726
    return AssetUploadTokens.insert({                                                                 // 727
      purpose: purpose,                                                                               // 728
      expires: new Date(Date.now() + 300000),  // in 5 minutes                                        // 729
    });                                                                                               // 730
  }                                                                                                   // 731
                                                                                                      // 732
  SandstormDb.prototype.fulfillAssetUpload = function (id) {                                          // 733
    // Indicates that the given asset upload has completed. It will be removed and its purpose        // 734
    // returned. If no matching upload exists, returns undefined.                                     // 735
                                                                                                      // 736
    check(id, String);                                                                                // 737
                                                                                                      // 738
    var upload = AssetUploadTokens.findAndModify({                                                    // 739
      query: {_id: id},                                                                               // 740
      remove: true                                                                                    // 741
    });                                                                                               // 742
                                                                                                      // 743
    if (upload.expires.valueOf() < Date.now()) {                                                      // 744
      return undefined;  // already expired                                                           // 745
    } else {                                                                                          // 746
      return upload.purpose;                                                                          // 747
    }                                                                                                 // 748
  }                                                                                                   // 749
                                                                                                      // 750
  function cleanupExpiredAssetUploads() {                                                             // 751
    AssetUploadTokens.remove({expires: {$lt: Date.now()}});                                           // 752
  }                                                                                                   // 753
                                                                                                      // 754
  // Cleanup tokens every hour, with a random phase.                                                  // 755
  Meteor.setTimeout(function () {                                                                     // 756
    cleanupExpiredAssetUploads();                                                                     // 757
    Meteor.setInterval(function () {                                                                  // 758
      cleanupExpiredAssetUploads();                                                                   // 759
    }, 3600000);                                                                                      // 760
  }, Math.floor(Math.random() * 3600000));                                                            // 761
}                                                                                                     // 762
                                                                                                      // 763
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/profile.js                                                                   //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                 // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                              // 2
// All rights reserved.                                                                               // 3
//                                                                                                    // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                    // 5
// you may not use this file except in compliance with the License.                                   // 6
// You may obtain a copy of the License at                                                            // 7
//                                                                                                    // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                       // 9
//                                                                                                    // 10
// Unless required by applicable law or agreed to in writing, software                                // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                  // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                           // 13
// See the License for the specific language governing permissions and                                // 14
// limitations under the License.                                                                     // 15
                                                                                                      // 16
var makeIdenticon;                                                                                    // 17
var httpProtocol;                                                                                     // 18
var sha256;                                                                                           // 19
                                                                                                      // 20
if (Meteor.isServer) {                                                                                // 21
  Meteor.publish("accountIdentities", function () {                                                   // 22
    if (!Meteor.userId) return [];                                                                    // 23
                                                                                                      // 24
    return [                                                                                          // 25
      Meteor.users.find(Meteor.userId,                                                                // 26
        {fields: {                                                                                    // 27
          "profile":1,                                                                                // 28
          "devName":1,                                                                                // 29
          "expires":1,                                                                                // 30
                                                                                                      // 31
          "services.google.id":1,                                                                     // 32
          "services.google.email":1,                                                                  // 33
          "services.google.verified_email":1,                                                         // 34
          "services.google.name":1,                                                                   // 35
          "services.google.picture":1,                                                                // 36
          "services.google.gender":1,                                                                 // 37
                                                                                                      // 38
          "services.github.id":1,                                                                     // 39
          "services.github.email":1,                                                                  // 40
          "services.github.username":1,                                                               // 41
                                                                                                      // 42
          "services.emailToken.email":1                                                               // 43
        }})                                                                                           // 44
    ];                                                                                                // 45
  });                                                                                                 // 46
                                                                                                      // 47
  makeIdenticon = function (id) {                                                                     // 48
    // We only make identicons client-side.                                                           // 49
    return undefined;                                                                                 // 50
  }                                                                                                   // 51
                                                                                                      // 52
  var Url = Npm.require("url");                                                                       // 53
  httpProtocol = Url.parse(process.env.ROOT_URL).protocol;                                            // 54
                                                                                                      // 55
  var Crypto = Npm.require("crypto");                                                                 // 56
  sha256 = function (data) {                                                                          // 57
    return Crypto.createHash("sha256").update(data).digest("hex");                                    // 58
  };                                                                                                  // 59
} else {                                                                                              // 60
  var identiconCache = {};                                                                            // 61
                                                                                                      // 62
  makeIdenticon = function (id) {                                                                     // 63
    // Given a cryptographic hash as input, generate an identicon. We always use the user's           // 64
    // identity ID as the input hash (the same thing passed to apps as the user ID).                  // 65
                                                                                                      // 66
    // Pass to identicon.js exactly the hash that we'd pass to apps as X-Sandstorm-User-Id, so        // 67
    // that apps can themselves use identicon.js to produce consistent identicons. As it turns out    // 68
    // identicon.js doesn't use the second half of the hash even if we provide it, but slice it       // 69
    // anyway to be safe.                                                                             // 70
    id = id.slice(0, 32);                                                                             // 71
                                                                                                      // 72
    if (id in identiconCache) {                                                                       // 73
      return identiconCache[id];                                                                      // 74
    }                                                                                                 // 75
                                                                                                      // 76
    // Unfortunately, Github's algorithm uses MD5. Whatever, we don't expect these to be secure.      // 77
    var data = new Identicon(id, 512).toString();                                                     // 78
    var result = "data:image/png;base64," + data;                                                     // 79
    identiconCache[id] = result;                                                                      // 80
    return result;                                                                                    // 81
  }                                                                                                   // 82
                                                                                                      // 83
  httpProtocol = window.location.protocol;                                                            // 84
                                                                                                      // 85
  sha256 = function (data) {                                                                          // 86
    return CryptoJS.SHA256(data).toString();                                                          // 87
  };                                                                                                  // 88
}                                                                                                     // 89
                                                                                                      // 90
var GENDERS = {male: "male", female: "female", neutral: "neutral", robot: "robot"};                   // 91
                                                                                                      // 92
function staticAssetUrl(id, staticHost) {                                                             // 93
  if (id) {                                                                                           // 94
    return staticHost + "/" + id;                                                                     // 95
  } else {                                                                                            // 96
    return undefined;                                                                                 // 97
  }                                                                                                   // 98
}                                                                                                     // 99
                                                                                                      // 100
function filterHandle(handle) {                                                                       // 101
  // Convert disallowed letters into underscores, but no more than one underscore in a row.           // 102
  return handle && handle.toLowerCase().split(/[^a-z0-9_]+/g).join("_");                              // 103
}                                                                                                     // 104
                                                                                                      // 105
function emailToHandle(email) {                                                                       // 106
  // Turn an email address into a handle.                                                             // 107
                                                                                                      // 108
  if (!email) return undefined;                                                                       // 109
                                                                                                      // 110
  // Use the stuff before the @. Ignore stuff after '+' because it's commonly used for filters        // 111
  // on the same account.                                                                             // 112
  var parts = email.split("@");                                                                       // 113
  var base = filterHandle(parts[0].split("+")[0]);                                                    // 114
                                                                                                      // 115
  var domain = (parts[1]||"").split(".");                                                             // 116
  if (domain[domain.length - 1] === "name") {                                                         // 117
    // Oh, a .name domain. Let's use                                                                  // 118
    base = filterHandle(domain.slice(0, domain.length - 1).join("."));                                // 119
  } else if (_.contains(["me", "self", "contact", "admin", "administrator", "root", "info",           // 120
                         "sandstorm", "sandstormio", "inbox", "indiegogo", "mail", "email"],          // 121
                         base)) {                                                                     // 122
    // This is probably an address at a vanity domain. Use the domain itself as the handle.           // 123
    base = filterHandle(domain[0]);                                                                   // 124
  }                                                                                                   // 125
                                                                                                      // 126
  return filterHandle(base);                                                                          // 127
}                                                                                                     // 128
                                                                                                      // 129
function identityId(service, id) {                                                                    // 130
  return sha256(service + ":" + id);                                                                  // 131
}                                                                                                     // 132
                                                                                                      // 133
function googleIdentity(user, staticHost) {                                                           // 134
  var google = (user.services && user.services.google) || {};                                         // 135
  var profile = user.profile || {};                                                                   // 136
  var id = identityId("google", google.id);                                                           // 137
                                                                                                      // 138
  return {                                                                                            // 139
    service: "google",                                                                                // 140
    id: id,                                                                                           // 141
    name: profile.name || google.name || "Name Unknown",                                              // 142
    handle: profile.handle || emailToHandle(google.email) ||                                          // 143
            filterHandle(profile.name || google.name) || "unknown",                                   // 144
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 145
    pronoun: profile.pronoun || GENDERS[google.gender] || "neutral",                                  // 146
  }                                                                                                   // 147
}                                                                                                     // 148
                                                                                                      // 149
function githubIdentity(user, staticHost) {                                                           // 150
  var github = (user.services && user.services.github) || {};                                         // 151
  var profile = user.profile || {};                                                                   // 152
  var id = identityId("github", github.id);                                                           // 153
                                                                                                      // 154
  return {                                                                                            // 155
    service: "github",                                                                                // 156
    id: id,                                                                                           // 157
    name: profile.name || github.username || "Name Unknown",                                          // 158
    handle: profile.handle || filterHandle(github.username) ||                                        // 159
            filterHandle(profile.name) || "unknown",                                                  // 160
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 161
    pronoun: profile.pronoun,                                                                         // 162
  }                                                                                                   // 163
}                                                                                                     // 164
                                                                                                      // 165
function emailIdentity(user, staticHost) {                                                            // 166
  var email = (user.services && user.services.emailToken) || {};                                      // 167
  var profile = user.profile || {};                                                                   // 168
  var id = identityId("email", email.email);                                                          // 169
                                                                                                      // 170
  return {                                                                                            // 171
    service: "email",                                                                                 // 172
    id: id,                                                                                           // 173
    name: profile.name || email.email.split("@")[0] || "Name Unknown",                                // 174
    handle: profile.handle || emailToHandle(email.email) ||                                           // 175
            filterHandle(profile.name) || "unknown",                                                  // 176
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 177
    pronoun: profile.pronoun,                                                                         // 178
  }                                                                                                   // 179
}                                                                                                     // 180
                                                                                                      // 181
function devIdentity(user, staticHost) {                                                              // 182
  var email = (user.services && user.services.emailToken) || {};                                      // 183
  var profile = user.profile || {};                                                                   // 184
  var name = user.devName.split(" ")[0].toLowerCase();                                                // 185
  var id = identityId("dev", user.devName);                                                           // 186
                                                                                                      // 187
  return {                                                                                            // 188
    service: "dev",                                                                                   // 189
    id: id,                                                                                           // 190
    name: profile.name || user.devName,                                                               // 191
    handle: profile.handle || filterHandle(name),                                                     // 192
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 193
    pronoun: profile.pronoun ||                                                                       // 194
             (_.contains(["alice", "carol", "eve"], name) ? "female" :                                // 195
              _.contains(["bob", "dave"], name) ? "male" : "neutral"),                                // 196
  }                                                                                                   // 197
}                                                                                                     // 198
                                                                                                      // 199
function demoIdentity(user, staticHost) {                                                             // 200
  var email = (user.services && user.services.emailToken) || {};                                      // 201
  var profile = user.profile || {};                                                                   // 202
  var id = identityId("demo", user._id);                                                              // 203
                                                                                                      // 204
  return {                                                                                            // 205
    service: "demo",                                                                                  // 206
    id: id,                                                                                           // 207
    name: profile.name || "Demo User",                                                                // 208
    handle: profile.handle || "demo",                                                                 // 209
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 210
    pronoun: profile.pronoun || "neutral",                                                            // 211
  }                                                                                                   // 212
}                                                                                                     // 213
                                                                                                      // 214
SandstormDb.getUserIdentities = function (user) {                                                     // 215
  // Given a user object, return all of the user's identities.                                        // 216
  //                                                                                                  // 217
  // On the client, must be subscribed "accountIdentities" for the user.                              // 218
                                                                                                      // 219
  var staticHost = httpProtocol + "//" + makeWildcardHost("static");                                  // 220
                                                                                                      // 221
  var result = [];                                                                                    // 222
  if (user && user.services) {                                                                        // 223
    if ("google" in user.services) {                                                                  // 224
      result.push(googleIdentity(user, staticHost));                                                  // 225
    }                                                                                                 // 226
    if ("github" in user.services) {                                                                  // 227
      result.push(githubIdentity(user, staticHost));                                                  // 228
    }                                                                                                 // 229
    if ("emailToken" in user.services) {                                                              // 230
      result.push(emailIdentity(user, staticHost));                                                   // 231
    }                                                                                                 // 232
    if ("devName" in user) {                                                                          // 233
      result.push(devIdentity(user, staticHost));                                                     // 234
    }                                                                                                 // 235
    if ("expires" in user) {                                                                          // 236
      result.push(demoIdentity(user, staticHost));                                                    // 237
    }                                                                                                 // 238
  }                                                                                                   // 239
  return result;                                                                                      // 240
}                                                                                                     // 241
                                                                                                      // 242
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/user.js                                                                      //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                 // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                              // 2
// All rights reserved.                                                                               // 3
//                                                                                                    // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                    // 5
// you may not use this file except in compliance with the License.                                   // 6
// You may obtain a copy of the License at                                                            // 7
//                                                                                                    // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                       // 9
//                                                                                                    // 10
// Unless required by applicable law or agreed to in writing, software                                // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                  // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                           // 13
// See the License for the specific language governing permissions and                                // 14
// limitations under the License.                                                                     // 15
                                                                                                      // 16
var Future = Npm.require("fibers/future");                                                            // 17
                                                                                                      // 18
userPictureUrl = function (user) {                                                                    // 19
  if (user.services) {                                                                                // 20
    // Try to determine user's avatar URL from login service.                                         // 21
                                                                                                      // 22
    var google = user.services.google;                                                                // 23
    if (!user.profile.picture && google && google.picture) {                                          // 24
      return google.picture;                                                                          // 25
    }                                                                                                 // 26
                                                                                                      // 27
    var github = user.services.github;                                                                // 28
    if (!user.profile.picture && github && github.id) {                                               // 29
      return "https://avatars.githubusercontent.com/u/" + github.id;                                  // 30
    }                                                                                                 // 31
                                                                                                      // 32
    // Note that we do NOT support Gravatar for email addresses because pinging Gravatar would be     // 33
    // a data leak, revealing that the user has logged into this Sandstorm server. Google and         // 34
    // Github are different because they are actually the identity providers, so they already know    // 35
    // the user logged in.                                                                            // 36
  }                                                                                                   // 37
}                                                                                                     // 38
                                                                                                      // 39
fetchPicture = function (url) {                                                                       // 40
  try {                                                                                               // 41
    var result = HTTP.get(url, {                                                                      // 42
      npmRequestOptions: { encoding: null },                                                          // 43
      timeout: 5000                                                                                   // 44
    });                                                                                               // 45
                                                                                                      // 46
    var metadata = {};                                                                                // 47
                                                                                                      // 48
    metadata.mimeType = result.headers["content-type"];                                               // 49
    if (metadata.mimeType.lastIndexOf("image/png", 0) === -1 &&                                       // 50
        metadata.mimeType.lastIndexOf("image/jpeg", 0) === -1) {                                      // 51
      throw new Error("unexpected Content-Type:", metadata.mimeType);                                 // 52
    }                                                                                                 // 53
    var enc = result.headers["content-encoding"];                                                     // 54
    if (enc && enc !== "identity") {                                                                  // 55
      metadata.encoding = enc;                                                                        // 56
    }                                                                                                 // 57
                                                                                                      // 58
    return addStaticAsset(metadata, result.content);                                                  // 59
  } catch (err) {                                                                                     // 60
    console.error("failed to fetch user profile picture:", url, err.stack);                           // 61
  }                                                                                                   // 62
}                                                                                                     // 63
                                                                                                      // 64
var ValidHandle = Match.Where(function (handle) {                                                     // 65
  check(handle, String);                                                                              // 66
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                        // 67
});                                                                                                   // 68
                                                                                                      // 69
Accounts.onCreateUser(function (options, user) {                                                      // 70
  // The first non-dev user to sign in should be automatically upgraded to admin.                     // 71
  // Dev users are identified by having the devName field.                                            // 72
  if (Meteor.users.find({devName: {$exists: 0}}).count() === 0 && !user.devName) {                    // 73
    user.isAdmin = true;                                                                              // 74
    user.signupKey = "admin";                                                                         // 75
  }                                                                                                   // 76
                                                                                                      // 77
  // Check profile.                                                                                   // 78
  if (options.profile) {                                                                              // 79
    // TODO(cleanup): This check also appears in accounts-ui-methods.js.                              // 80
    check(options.profile, Match.ObjectIncluding({                                                    // 81
      name: Match.Optional(String),                                                                   // 82
      handle: Match.Optional(ValidHandle),                                                            // 83
      pronoun: Match.Optional(Match.OneOf("male", "female", "neutral", "robot")),                     // 84
    }));                                                                                              // 85
                                                                                                      // 86
    user.profile = options.profile;                                                                   // 87
  } else {                                                                                            // 88
    user.profile = {};                                                                                // 89
  }                                                                                                   // 90
                                                                                                      // 91
  // Try downloading avatar.                                                                          // 92
  var url = userPictureUrl(user);                                                                     // 93
  if (url) {                                                                                          // 94
    var assetId = fetchPicture(url);                                                                  // 95
    if (assetId) {                                                                                    // 96
      user.profile.picture = assetId;                                                                 // 97
    }                                                                                                 // 98
  }                                                                                                   // 99
                                                                                                      // 100
  return user;                                                                                        // 101
});                                                                                                   // 102
                                                                                                      // 103
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/migrations.js                                                                //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Minimal tooling for doing run-at-least-once, ordered migrations.                                   // 1
//                                                                                                    // 2
// Because migrations can experience partial failure and likely have                                  // 3
// side-effects, we should be careful to make sure all migrations are                                 // 4
// idempotent and safe to accidentally run multiple times.                                            // 5
                                                                                                      // 6
var updateLoginStyleToRedirect = function() {                                                         // 7
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;          // 8
  ["google", "github"].forEach(function(serviceName) {                                                // 9
    var config = configurations.findOne({service: serviceName});                                      // 10
    if (config && config.loginStyle !== "redirect") {                                                 // 11
      configurations.update({service: serviceName}, {$set: {loginStyle: "redirect"}});                // 12
    }                                                                                                 // 13
  });                                                                                                 // 14
};                                                                                                    // 15
                                                                                                      // 16
var enableLegacyOAuthProvidersIfNotInSettings = function() {                                          // 17
  // In the before time, Google and Github login were enabled by default.                             // 18
  //                                                                                                  // 19
  // This actually didn't make much sense, required the first user to configure                       // 20
  // OAuth, and had some trust-the-first-user properties that weren't totally                         // 21
  // secure.                                                                                          // 22
  //                                                                                                  // 23
  // Now, we have admin-token, so we wish to disable all logins by default                            // 24
  // (since they need to be configured anyway) but since users may have never                         // 25
  // explicitly told Sandstorm that Google or Github login should be enabled,                         // 26
  // we can't just use the value in the Settings collection, since it might                           // 27
  // never have been set.                                                                             // 28
  //                                                                                                  // 29
  // Thus, if the service is configured but absent from Settings, we should                           // 30
  // explicitly enable it in Settings, and then the rest of the logic can just                        // 31
  // depend on what value is in Settings and default to false without breaking                        // 32
  // user installations.                                                                              // 33
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;          // 34
  ["google", "github"].forEach(function(serviceName) {                                                // 35
    var config = configurations.findOne({service: serviceName});                                      // 36
    var serviceConfig = Settings.findOne({_id: serviceName});                                         // 37
    if (config && !serviceConfig) {                                                                   // 38
      // Only explicitly enable the login service if:                                                 // 39
      // 1) the service is already configured                                                         // 40
      // 2) there is no sandstorm configuration already present (the user was                         // 41
      //    using the previous default behavior).                                                     // 42
      Settings.insert({_id: serviceName, value: true});                                               // 43
    }                                                                                                 // 44
  });                                                                                                 // 45
};                                                                                                    // 46
                                                                                                      // 47
var denormalizeInviteInfo = function() {                                                              // 48
  // When a user is invited via a signup token, the `signupKey` field of their user table entry       // 49
  // has always been populated to indicate the key they used. This points into the SignupKeys table   // 50
  // which has more information about the key, namely a freeform note entered by the admin when       // 51
  // they created the key. In the case that the email invite form was used, the note has the form     // 52
  // "E-mail invite to <address>".                                                                    // 53
  //                                                                                                  // 54
  // Later, we decided it was useful to indicate in the users table visible to the admin              // 55
  // information about the invite terms. Namely, for email invites we want to show the address        // 56
  // and for others we want to show the note. To make this efficient, fields `signupNote` and         // 57
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing       // 58
  // from the SignupKeys table.                                                                       // 59
                                                                                                      // 60
  Meteor.users.find().forEach(function (user) {                                                       // 61
    if (user.signupKey && (typeof user.signupKey) === "string" && user.signupKey !== "admin") {       // 62
      var signupInfo = SignupKeys.findOne(user.signupKey);                                            // 63
      if (signupInfo && signupInfo.note) {                                                            // 64
        var newFields = { signupNote: signupInfo.note };                                              // 65
                                                                                                      // 66
        var prefix = "E-mail invite to ";                                                             // 67
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                              // 68
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                               // 69
        }                                                                                             // 70
                                                                                                      // 71
        Meteor.users.update(user._id, {$set: newFields});                                             // 72
      }                                                                                               // 73
    }                                                                                                 // 74
  });                                                                                                 // 75
}                                                                                                     // 76
                                                                                                      // 77
function mergeRoleAssignmentsIntoApiTokens() {                                                        // 78
  RoleAssignments.find().forEach(function(roleAssignment) {                                           // 79
    ApiTokens.insert({                                                                                // 80
      grainId: roleAssignment.grainId,                                                                // 81
      userId: roleAssignment.sharer,                                                                  // 82
      roleAssignment: roleAssignment.roleAssignment,                                                  // 83
      petname: roleAssignment.petname,                                                                // 84
      created: roleAssignment.created,                                                                // 85
      owner: {user: {userId: roleAssignment.recipient,                                                // 86
                     title: roleAssignment.title}},                                                   // 87
    });                                                                                               // 88
  });                                                                                                 // 89
}                                                                                                     // 90
                                                                                                      // 91
function fixOasisStorageUsageStats() {                                                                // 92
  // Before storage quota enforcement was first implemented, there were some bugs with the way the    // 93
  // Oasis backend tracked storage usage, which were not initially noticed since the resulting        // 94
  // number wasn't acutally used for anything.                                                        // 95
  //                                                                                                  // 96
  // With the underlying bugs fixed, simply starting up each grain (supervisor only, not the app)     // 97
  // and then immediately shutting it down should cause the sizes to be recalculated correctly.       // 98
  //                                                                                                  // 99
  // This problem only applies to Oasis.                                                              // 100
                                                                                                      // 101
  if (Meteor.settings.public.quotaEnabled) {                                                          // 102
    var dummyCommand = {                                                                              // 103
      argv: ["foo", "bar"],                                                                           // 104
      environ: []                                                                                     // 105
    };                                                                                                // 106
                                                                                                      // 107
    Grains.find().forEach(function (grain) {                                                          // 108
      try {                                                                                           // 109
        console.log("refreshing grain:", grain._id);                                                  // 110
                                                                                                      // 111
        // We take a highly synchronous approach here so that the logging is understandable.          // 112
        var supervisor = waitPromise(sandstormBackend.startGrain(                                     // 113
            grain.userId, grain._id, grain.packageId, dummyCommand, false, false)).supervisor;        // 114
        try {                                                                                         // 115
          waitPromise(supervisor.shutdown());                                                         // 116
          throw new Error("shutdown() didn't disconnect!");                                           // 117
        } catch (err) {                                                                               // 118
          if (err.kjType !== "disconnected") throw err;                                               // 119
        }                                                                                             // 120
      } catch (err) {                                                                                 // 121
        console.log(err.stack);                                                                       // 122
      }                                                                                               // 123
    });                                                                                               // 124
  }                                                                                                   // 125
}                                                                                                     // 126
                                                                                                      // 127
function fetchProfilePictures() {                                                                     // 128
  Meteor.users.find({}).forEach(function (user) {                                                     // 129
    var url = userPictureUrl(user);                                                                   // 130
    if (url) {                                                                                        // 131
      console.log("Fetching user picture:", url);                                                     // 132
      var assetId = fetchPicture(url);                                                                // 133
      if (assetId) {                                                                                  // 134
        Meteor.users.update(user._id, {$set: {"profile.picture": assetId}});                          // 135
      }                                                                                               // 136
    }                                                                                                 // 137
  });                                                                                                 // 138
}                                                                                                     // 139
                                                                                                      // 140
// This must come after all the functions named within are defined.                                   // 141
// Only append to this list!  Do not modify or remove list entries;                                   // 142
// doing so is likely change the meaning and semantics of user databases.                             // 143
var MIGRATIONS = [                                                                                    // 144
  updateLoginStyleToRedirect,                                                                         // 145
  enableLegacyOAuthProvidersIfNotInSettings,                                                          // 146
  denormalizeInviteInfo,                                                                              // 147
  mergeRoleAssignmentsIntoApiTokens,                                                                  // 148
  fixOasisStorageUsageStats,                                                                          // 149
  fetchProfilePictures,                                                                               // 150
];                                                                                                    // 151
                                                                                                      // 152
function migrateToLatest() {                                                                          // 153
  var applied = Migrations.findOne({_id: "migrations_applied"});                                      // 154
  var start;                                                                                          // 155
  if (!applied) {                                                                                     // 156
    // Migrations table is not yet seeded with a value.  This means it has                            // 157
    // applied 0 migrations.  Persist this.                                                           // 158
    Migrations.insert({_id: "migrations_applied", value: 0});                                         // 159
    start = 0;                                                                                        // 160
  } else {                                                                                            // 161
    start = applied.value;                                                                            // 162
  }                                                                                                   // 163
  console.log("Migrations applied: " + start + "/" + MIGRATIONS.length);                              // 164
                                                                                                      // 165
  for (var i = start ; i < MIGRATIONS.length ; i++) {                                                 // 166
    // Apply migration i, then record that migration i was successfully run.                          // 167
    console.log("Applying migration " + (i+1));                                                       // 168
    MIGRATIONS[i]();                                                                                  // 169
    Migrations.update({_id: "migrations_applied"}, {$set: {value: i+1}});                             // 170
    console.log("Applied migration " + (i+1));                                                        // 171
  }                                                                                                   // 172
}                                                                                                     // 173
                                                                                                      // 174
// Apply all migrations on startup.                                                                   // 175
Meteor.startup(migrateToLatest);                                                                      // 176
                                                                                                      // 177
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-db'] = {
  SandstormDb: SandstormDb
};

})();

//# sourceMappingURL=sandstorm-db.js.map
